////////////////////////////////////////
// Programming project code
// UniBw M, 2022, 2023, 2024
// www.unibw.de/inf2
// (c) Mark Minas (mark.minas@unibw.de)
////////////////////////////////////////

package pp.battleship.client.gui;


import static java.util.Objects.requireNonNull;

import com.jme3.effect.ParticleEmitter;
import com.jme3.effect.ParticleMesh.Type;
import com.jme3.effect.shapes.EmitterSphereShape;
import com.jme3.material.Material;
import com.jme3.material.RenderState.BlendMode;
import com.jme3.math.ColorRGBA;
import com.jme3.math.FastMath;
import com.jme3.math.Vector3f;
import com.jme3.renderer.queue.RenderQueue.ShadowMode;
import com.jme3.scene.Geometry;
import com.jme3.scene.Node;
import com.jme3.scene.Spatial;
import com.jme3.scene.shape.Box;
import com.jme3.scene.shape.Cylinder;

import pp.battleship.client.BattleshipApp;
import pp.battleship.model.Battleship;
import pp.battleship.model.Rotation;
import pp.battleship.model.ShipMap;
import pp.battleship.model.Shot;
import static pp.util.FloatMath.HALF_PI;
import static pp.util.FloatMath.PI;

/**
 * The {@code SeaSynchronizer} class is responsible for synchronizing the graphical
 * representation of the ships and shots on the sea map with the underlying data model.
 * It extends the {@link ShipMapSynchronizer} to provide specific synchronization
 * logic for the sea map.
 */
class SeaSynchronizer extends ShipMapSynchronizer {
    private static final String UNSHADED = "Common/MatDefs/Misc/Unshaded.j3md"; //NON-NLS
    private static final String KING_GEORGE_V_MODEL = "Models/Transporter/cr90.j3o"; //NON-NLS
    private static final String TIE_FIGHTER = "Models/TieFighter/TieFighter.j3o"; //NON-NLS
    private static final String TRANSPORTER = "Models/KingGeorgeV/KingGeorge.j3o"; //NON-NLS
    //private static final String VENATOR = "Models/Venator/Venator.j3o"; //NON-NLS
    private static final String WING = "Models/Wing/Wing.j3o"; //NON-NLS
    
    private static final String COLOR = "Color"; //NON-NLS
    private static final String SHIP = "ship"; //NON-NLS
    private static final String SHOT = "shot"; //NON-NLS
    private static final ColorRGBA BOX_COLOR = ColorRGBA.Gray;
    private static final ColorRGBA SPLASH_COLOR = new ColorRGBA(0f, 0f, 1f, 0.4f);
    private static final ColorRGBA HIT_COLOR = new ColorRGBA(1f, 0f, 0f, 0.4f);

    private final ShipMap map;
    private final BattleshipApp app;

    private ParticleEmitter flame, flash, spark, roundspark, smoketrail, debris,
            shockwave;


    private static final int COUNT_FACTOR = 1;
    private static final float COUNT_FACTOR_F = 1f;

    private static final boolean POINT_SPRITE = true;
    private static final Type EMITTER_TYPE = POINT_SPRITE ? Type.Point : Type.Triangle;
    /**
     * Creates a Flame texture with indefininite Lifetime
     * @return flame texture
     */

     private ParticleEmitter createFlame() {
        flame = new ParticleEmitter("Flame", EMITTER_TYPE, 32 * COUNT_FACTOR);
        flame.setSelectRandomImage(true);
        flame.setStartColor(new ColorRGBA(1f, 0.4f, 0.05f, (1f / COUNT_FACTOR_F)));
        flame.setEndColor(new ColorRGBA(.4f, .22f, .12f, 0f));
        flame.setStartSize(0.1f);
        flame.setEndSize(0.5f);
        flame.setShape(new EmitterSphereShape(Vector3f.ZERO, 1f));
        flame.setParticlesPerSec(0);
        flame.setGravity(0, -5, 0);
        flame.setLowLife(.4f);
        flame.setHighLife(.5f);
        flame.getParticleInfluencer().setInitialVelocity(new Vector3f(0, 7, 0));
        flame.getParticleInfluencer().setVelocityVariation(1f);
        flame.setImagesX(2);
        flame.setImagesY(2);
        Material mat = new Material(app.getAssetManager(), "Common/MatDefs/Misc/Particle.j3md");
        mat.setTexture("Texture", app.getAssetManager().loadTexture("Effects/Explosion/flame.png"));
        mat.setBoolean("PointSprite", POINT_SPRITE);
        flame.setMaterial(mat);
    
        // Setze die Höhe auf 2f
        flame.setLocalTranslation(0, 2f, 0);
    
        return flame;
    }

    private ParticleEmitter createSpark() {
        spark = new ParticleEmitter("Spark", Type.Triangle, 30 * COUNT_FACTOR);
        spark.setStartColor(new ColorRGBA(1f, 0.8f, 0.36f, 1.0f / COUNT_FACTOR_F));
        spark.setEndColor(new ColorRGBA(1f, 0.8f, 0.36f, 0f));
        spark.setStartSize(.5f);
        spark.setEndSize(.5f);
        spark.setFacingVelocity(true);
        spark.setParticlesPerSec(0);
        spark.setGravity(0, 5, 0);
        spark.setLowLife(1.1f);
        spark.setHighLife(1.5f);
        spark.getParticleInfluencer().setInitialVelocity(new Vector3f(0, 20, 0));
        spark.getParticleInfluencer().setVelocityVariation(1);
        spark.setImagesX(1);
        spark.setImagesY(1);
        Material mat = new Material(app.getAssetManager(), "Common/MatDefs/Misc/Particle.j3md");
        mat.setTexture("Texture", app.getAssetManager().loadTexture("Effects/Explosion/spark.png"));
        spark.setMaterial(mat);
    
        // Setze die Höhe auf 2f
        spark.setLocalTranslation(0, 2f, 0);
    
        return spark;
    }

    private ParticleEmitter createRoundSpark() {
        roundspark = new ParticleEmitter("RoundSpark", EMITTER_TYPE, 20 * COUNT_FACTOR);
        roundspark.setStartColor(new ColorRGBA(1f, 0.29f, 0.34f, (float) (1.0 / COUNT_FACTOR_F)));
        roundspark.setEndColor(new ColorRGBA(0, 0, 0, 0.5f / COUNT_FACTOR_F));
        roundspark.setStartSize(0.2f);
        roundspark.setEndSize(0.8f);
        roundspark.setShape(new EmitterSphereShape(Vector3f.ZERO, 1f));
        roundspark.setParticlesPerSec(0);
        roundspark.setGravity(0, -.5f, 0);
        roundspark.setLowLife(1.8f);
        roundspark.setHighLife(2f);
        roundspark.getParticleInfluencer().setInitialVelocity(new Vector3f(0, 3, 0));
        roundspark.getParticleInfluencer().setVelocityVariation(.5f);
        roundspark.setImagesX(1);
        roundspark.setImagesY(1);
        Material mat = new Material(app.getAssetManager(), "Common/MatDefs/Misc/Particle.j3md");
        mat.setTexture("Texture", app.getAssetManager().loadTexture("Effects/Explosion/roundspark.png"));
        mat.setBoolean("PointSprite", POINT_SPRITE);
        roundspark.setMaterial(mat);
    
        // Setze die Höhe auf 2f
        roundspark.setLocalTranslation(0, 2f, 0);
    
        return roundspark;
    }

    

    private ParticleEmitter createSmokeTrail() {
        smoketrail = new ParticleEmitter("SmokeTrail", Type.Triangle, 22 * COUNT_FACTOR);
        smoketrail.setStartColor(new ColorRGBA(1f, 0.8f, 0.36f, 1.0f / COUNT_FACTOR_F));
        smoketrail.setEndColor(new ColorRGBA(1f, 0.8f, 0.36f, 0f));
        smoketrail.setStartSize(.2f);
        smoketrail.setEndSize(1f);
        smoketrail.setFacingVelocity(true);
        smoketrail.setParticlesPerSec(0);
        smoketrail.setGravity(0, 1, 0);
        smoketrail.setLowLife(.4f);
        smoketrail.setHighLife(.5f);
        smoketrail.getParticleInfluencer().setInitialVelocity(new Vector3f(0, 12, 0));
        smoketrail.getParticleInfluencer().setVelocityVariation(1);
        smoketrail.setImagesX(1);
        smoketrail.setImagesY(3);
        Material mat = new Material(app.getAssetManager(), "Common/MatDefs/Misc/Particle.j3md");
        mat.setTexture("Texture", app.getAssetManager().loadTexture("Effects/Explosion/smoketrail.png"));
        smoketrail.setMaterial(mat);
    
        // Setze die Höhe auf 2f
        smoketrail.setLocalTranslation(0, 2f, 0);
    
        return smoketrail;
    }
    

    private ParticleEmitter createDebris() {
        debris = new ParticleEmitter("Debris", Type.Triangle, 15 * COUNT_FACTOR);
        debris.setSelectRandomImage(true);
        debris.setRandomAngle(true);
        debris.setRotateSpeed(FastMath.TWO_PI * 4);
        debris.setStartColor(new ColorRGBA(1f, 0.59f, 0.28f, 1.0f / COUNT_FACTOR_F));
        debris.setEndColor(new ColorRGBA(.5f, 0.5f, 0.5f, 0f));
        debris.setStartSize(.10f);
        debris.setEndSize(.15f);
        debris.setParticlesPerSec(0);
        debris.setGravity(0, 12f, 0);
        debris.setLowLife(1.4f);
        debris.setHighLife(1.5f);
        debris.getParticleInfluencer().setInitialVelocity(new Vector3f(0, 15, 0));
        debris.getParticleInfluencer().setVelocityVariation(.60f);
        debris.setImagesX(3);
        debris.setImagesY(3);
        Material mat = new Material(app.getAssetManager(), "Common/MatDefs/Misc/Particle.j3md");
        mat.setTexture("Texture", app.getAssetManager().loadTexture("Effects/Explosion/Debris.png"));
        debris.setMaterial(mat);
    
        // Setze die Höhe auf 2f
        debris.setLocalTranslation(0, 2f, 0);
    
        return debris;
    }
    

    private ParticleEmitter createShockwave(){
        shockwave = new ParticleEmitter("Shockwave", Type.Triangle, 1 * COUNT_FACTOR);
//        shockwave.setRandomAngle(true);
        shockwave.setFaceNormal(Vector3f.UNIT_Y);
        shockwave.setStartColor(new ColorRGBA(.48f, 0.17f, 0.01f, .8f / COUNT_FACTOR_F));
        shockwave.setEndColor(new ColorRGBA(.48f, 0.17f, 0.01f, 0f));

        shockwave.setStartSize(0f);
        shockwave.setEndSize(3f);

        shockwave.setParticlesPerSec(0);
        shockwave.setGravity(0, 0, 0);
        shockwave.setLowLife(0.5f);
        shockwave.setHighLife(0.5f);
        shockwave.getParticleInfluencer()
                 .setInitialVelocity(new Vector3f(0, 0, 0));
        shockwave.getParticleInfluencer().setVelocityVariation(0f);
        shockwave.setImagesX(1);
        shockwave.setImagesY(1);
        Material mat = new Material(app.getAssetManager(), "Common/MatDefs/Misc/Particle.j3md");
        mat.setTexture("Texture", app.getAssetManager().loadTexture("Effects/Explosion/shockwave.png"));
        shockwave.setMaterial(mat);

        return shockwave;
    }

    private ParticleEmitter createMovingSmokeEmitter() {
        ParticleEmitter smokeEmitter = new ParticleEmitter("SmokeEmitter", Type.Triangle, 300);
        smokeEmitter.setGravity(0, 0, 0);
        smokeEmitter.getParticleInfluencer().setVelocityVariation(1);
        smokeEmitter.setLowLife(1);
        smokeEmitter.setHighLife(1);
        smokeEmitter.getParticleInfluencer().setInitialVelocity(new Vector3f(0, .5f, 0));
        smokeEmitter.setImagesX(15);
        Material mat = new Material(app.getAssetManager(), "Common/MatDefs/Misc/Particle.j3md");
        mat.setTexture("Texture", app.getAssetManager().loadTexture("Effects/Smoke/Smoke.png"));
        smokeEmitter.setMaterial(mat);
    
        // Setze die Höhe auf 2f
        smokeEmitter.setLocalTranslation(0, 2f, 0);
    
        return smokeEmitter;
    }
    

    /**
     * Constructs a {@code SeaSynchronizer} object with the specified application, root node, and ship map.
     *
     * @param app  the Battleship application
     * @param root the root node to which graphical elements will be attached
     * @param map  the ship map containing the ships and shots
     */
    public SeaSynchronizer(BattleshipApp app, Node root, ShipMap map) {
        super(app.getGameLogic().getOwnMap(), root);
        this.app = app;
        this.map = map;
        addExisting();
    }

    /**
     * Visits a {@link Shot} and creates a graphical representation of it.
     * If the shot is a hit, it attaches the representation to the ship node.
     *
     * @param shot the shot to be represented
     * @return the graphical representation of the shot, or null if the shot is a hit
     * and the representation has been attached to the ship node
     */
    @Override
    public Spatial visit(Shot shot) {
        return shot.isHit() ? handleHit(shot) : createCylinder(shot);
    }

    /**
     * Handles a hit by attaching its representation to the node that
     * contains the ship model as a child so that it moves with the ship.
     *
     * @param shot a hit
     * @return always null to prevent the representation from being attached
     * to the items node as well
     */
    private Spatial handleHit(Shot shot) {
        final Battleship ship = requireNonNull(map.findShipAt(shot), "Missing ship");
        final Node shipNode = requireNonNull((Node) getSpatial(ship), "Missing ship node");

       //final Geometry representation = createCylinder(shot);
       //representation.getLocalTranslation().subtractLocal(shipNode.getLocalTranslation());
       //shipNode.attachChild(representation);
        // Create a new node specifically for the hit effects
        Node hitEffectNode = new Node("HitEffectNode");

        // Create particle effects
        ParticleEmitter flame = createFlame();
       // ParticleEmitter flash = createFlash();
        ParticleEmitter spark = createSpark();
        ParticleEmitter roundSpark = createRoundSpark();
        ParticleEmitter smokeTrail = createSmokeTrail();
        ParticleEmitter debris = createDebris();
        ParticleEmitter shockwave = createShockwave();
        ParticleEmitter movingSmoke = createMovingSmokeEmitter(); // New moving smoke emitter

        // Attach all effects to the hitEffectNode
        hitEffectNode.attachChild(flame);
        hitEffectNode.attachChild(flash);
        hitEffectNode.attachChild(spark);
        hitEffectNode.attachChild(roundSpark);
        hitEffectNode.attachChild(smokeTrail);
        hitEffectNode.attachChild(debris);
        hitEffectNode.attachChild(shockwave);
        hitEffectNode.attachChild(movingSmoke);

        // Set the local translation for the hit effect to the point of impact
        hitEffectNode.setLocalTranslation(shot.getY() + 0.5f - shipNode.getLocalTranslation().x,
                                          0.5f, // Adjust as needed for height above the ship
                                          shot.getX() + 0.5f - shipNode.getLocalTranslation().z);

        // Attach the hitEffectNode to the shipNode so it moves with the ship
        shipNode.attachChild(hitEffectNode);

        // Emit particles when the hit happens
        flash.emitAllParticles();
        spark.emitAllParticles();
        smokeTrail.emitAllParticles();
        debris.emitAllParticles();
        shockwave.emitAllParticles();
        flame.emitAllParticles();
        roundSpark.emitAllParticles();



       return null;
    }

    /**
     * Creates a cylinder geometry representing the specified shot.
     * The appearance of the cylinder depends on whether the shot is a hit or a miss.
     *
     * @param shot the shot to be represented
     * @return the geometry representing the shot
     */
    private Geometry createCylinder(Shot shot) {
        final ColorRGBA color = shot.isHit() ? HIT_COLOR : SPLASH_COLOR;
        final float height = shot.isHit() ? 1.2f : 0.1f;
    
        final Cylinder cylinder = new Cylinder(2, 20, 0.45f, height, true);
        final Geometry geometry = new Geometry(SHOT, cylinder);
    
        geometry.setMaterial(createColoredMaterial(color));
        geometry.rotate(HALF_PI, 0f, 0f);
        
        // compute the center of the shot in world coordinates, with an additional 2f height
        geometry.setLocalTranslation(shot.getY() + 0.5f, 2f + height / 2, shot.getX() + 0.5f);
    
        return geometry;
    }

    /**
     * Visits a {@link Battleship} and creates a graphical representation of it.
     * The representation is either a 3D model or a simple box depending on the
     * type of battleship.
     *
     * @param ship the battleship to be represented
     * @return the node containing the graphical representation of the battleship
     */
    @Override
    public Spatial visit(Battleship ship) {
        final Node node = new Node(SHIP);
        node.attachChild(createShip(ship));
        // compute the center of the ship in world coordinates
        final float x = 0.5f * (ship.getMinY() + ship.getMaxY() + 1f);
        final float z = 0.5f * (ship.getMinX() + ship.getMaxX() + 1f);
        node.setLocalTranslation(x, 0f, z);
        node.addControl(new ShipControl(ship));
        return node;
    }

    /**
     * Creates the appropriate graphical representation of the specified battleship.
     * The representation is either a detailed model or a simple box based on the length of the ship.
     *
     * @param ship the battleship to be represented
     * @return the spatial representing the battleship
     */
    private Spatial createShip(Battleship ship) {
        if (ship.getLength() == 1) {
            return createTie_Fighter(ship);
        } else if (ship.getLength() == 2) {
            return createWing(ship);
        } else if (ship.getLength() == 3) {
            return createTransporter(ship);
        } else if (ship.getLength() == 4) {
            return createBattleship(ship);
        } else {
            return createBox(ship);
        }
    
    }
    

    /**
     * Creates a simple box to represent a battleship that is not of the "King George V" type.
     *
     * @param ship the battleship to be represented
     * @return the geometry representing the battleship as a box
     */
    private Spatial createBox(Battleship ship) {
        final Box box = new Box(0.5f * (ship.getMaxY() - ship.getMinY()) + 0.3f,
                                0.3f,
                                0.5f * (ship.getMaxX() - ship.getMinX()) + 0.3f);
        final Geometry geometry = new Geometry(SHIP, box);
        geometry.setMaterial(createColoredMaterial(BOX_COLOR));
        geometry.setShadowMode(ShadowMode.CastAndReceive);

        return geometry;
    }

    /**
     * Creates a new {@link Material} with the specified color.
     * If the color includes transparency (i.e., alpha value less than 1),
     * the material's render state is set to use alpha blending, allowing for
     * semi-transparent rendering.
     *
     * @param color the {@link ColorRGBA} to be applied to the material. If the alpha value
     *              of the color is less than 1, the material will support transparency.
     * @return a {@link Material} instance configured with the specified color and,
     * if necessary, alpha blending enabled.
     */
    private Material createColoredMaterial(ColorRGBA color) {
        final Material material = new Material(app.getAssetManager(), UNSHADED);
        if (color.getAlpha() < 1f)
            material.getAdditionalRenderState().setBlendMode(BlendMode.Alpha);
        material.setColor(COLOR, color);
        return material;
    }

    /**
     * Creates a detailed 3D model to represent all battleships.
     *
     * @param ship the battleship to be represented
     * @return the spatial representing the "King George V" battleship
     */
    private Spatial createBattleship(Battleship ship) {
        final Spatial model = app.getAssetManager().loadModel(KING_GEORGE_V_MODEL);
        

    // Berechne den Rotationswinkel
    float rotationAngle = calculateRotationAngle(ship.getRot())+HALF_PI;
    model.setLocalTranslation(new Vector3f(model.getLocalTranslation().x, 2f, model.getLocalTranslation().z));
    model.rotate(0f, rotationAngle, -HALF_PI);
    model.scale(1.2f);
    model.setShadowMode(ShadowMode.CastAndReceive);

    // Bestimme die Verschiebung basierend auf der Rotation
    Vector3f translation = model.getLocalTranslation();
    
    // Bei 0 oder 180 Grad wird auf der Z-Achse verschoben, bei 90 oder 270 Grad auf der X-Achse
    if (rotationAngle == 0 || rotationAngle == FastMath.PI) {
        // Verschiebe auf der Z-Achse und um 2f nach oben
        translation = translation.subtract(0,0,1);  // Nach hinten auf der Z-Achse, und um 2f nach oben
    } else if (rotationAngle == FastMath.HALF_PI || rotationAngle == 3 * FastMath.HALF_PI) {
        // Verschiebe auf der X-Achse und um 2f nach oben
        translation = translation.subtract(1,0,0);  // Nach hinten auf der X-Achse, und um 2f nach oben
    }
    
    model.setLocalTranslation(translation);
    
    return model;
    }
    private Spatial createTransporter(Battleship ship) {
        final Spatial model = app.getAssetManager().loadModel(TRANSPORTER);
        float rotationAngle = calculateRotationAngle(ship.getRot())+HALF_PI;

        model.rotate(-HALF_PI, calculateRotationAngle(ship.getRot())+ PI, 0f);
        model.scale(0.008f);
        model.setShadowMode(ShadowMode.CastAndReceive);
        model.setLocalTranslation(new Vector3f(model.getLocalTranslation().x, 2f, model.getLocalTranslation().z));
        Vector3f translation = model.getLocalTranslation();
        // Bei 0 oder 180 Grad wird auf der Z-Achse verschoben, bei 90 oder 270 Grad auf der X-Achse
        if (rotationAngle == 0 || rotationAngle == FastMath.PI) {
        // Verschiebe auf der Z-Achse und um 2f nach oben
        translation = translation.add(0,0,1f);  // Nach hinten auf der Z-Achse, und um 2f nach oben
        } else if (rotationAngle == FastMath.HALF_PI || rotationAngle == 3 * FastMath.HALF_PI) {
        // Verschiebe auf der X-Achse und um 2f nach oben
        translation = translation.add(1f,0,0);  // Nach hinten auf der X-Achse, und um 2f nach oben
      }
    
        model.setLocalTranslation(translation);
    
    
        return model;
    }
    private Spatial createWing(Battleship ship) {
    final Spatial model = app.getAssetManager().loadModel(WING);
    model.setLocalTranslation(new Vector3f(model.getLocalTranslation().x, 2f, model.getLocalTranslation().z));

    // Berechne den Rotationswinkel
    float rotationAngle = calculateRotationAngle(ship.getRot())+HALF_PI;
    model.rotate(0f, rotationAngle, 0f);
    model.scale(0.0008f);
    model.setShadowMode(ShadowMode.CastAndReceive);

    // Bestimme die Verschiebung basierend auf der Rotation
    Vector3f translation = model.getLocalTranslation();
    
    // Bei 0 oder 180 Grad wird auf der Z-Achse verschoben, bei 90 oder 270 Grad auf der X-Achse
    if (rotationAngle == 0 || rotationAngle == FastMath.PI) {
        // Verschiebe auf der Z-Achse und um 2f nach oben
        translation = translation.add(0,0,0);  // Nach hinten auf der Z-Achse, und um 2f nach oben
    } else if (rotationAngle == FastMath.HALF_PI || rotationAngle == 3 * FastMath.HALF_PI) {
        // Verschiebe auf der X-Achse und um 2f nach oben
        translation = translation.add(0,0,0);  // Nach hinten auf der X-Achse, und um 2f nach oben
    }
    
    model.setLocalTranslation(translation);
    
    return model;
    
}
    private Spatial createTie_Fighter(Battleship ship) {
        final Spatial model = app.getAssetManager().loadModel(TIE_FIGHTER);

        model.rotate(-0f, calculateRotationAngle(ship.getRot()), 0f);
        model.scale(0.07f);
        model.setShadowMode(ShadowMode.CastAndReceive);
        model.setLocalTranslation(new Vector3f(model.getLocalTranslation().x, 2f, model.getLocalTranslation().z));

        return model;
    }
    

    /**
     * Calculates the rotation angle for the specified rotation.
     *
     * @param rot the rotation of the battleship
     * @return the rotation angle in radians
     */
    private static float calculateRotationAngle(Rotation rot) {
        return switch (rot) {
            case RIGHT -> HALF_PI;
            case DOWN -> 0f;
            case LEFT -> -HALF_PI;
            case UP -> PI;
        };
    }
}
